/**
    bespoke synth, a software modular synthesizer
    Copyright (C) 2021 Ryan Challinor (contact: awwbees@gmail.com)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
**/
//
//  PerformanceTimer.h
//  Bespoke
//
//  Created by Ryan Challinor on 3/26/15.
//
//

#pragma once

#include "OpenFrameworksPort.h"

class PerformanceTimer;

class TimerInstance
{
public:
   TimerInstance(std::string name, PerformanceTimer& manager);
   ~TimerInstance();

private:
   long mTimerStart;
   std::string mName;
   PerformanceTimer& mManager;
};

class PerformanceTimer
{
public:
   void RecordCost(std::string name, long cost);
   void PrintCosts();

private:
   struct Cost
   {
      Cost(std::string name, long cost)
      : mName(name)
      , mCost(cost)
      {}
      std::string mName;
      long mCost;
   };

   static bool SortCosts(const PerformanceTimer::Cost& a, const PerformanceTimer::Cost& b);

   std::vector<Cost> mCostTable;
};
